<?php

namespace Drupal\ostickets\Plugin\WebformHandler;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Serialization\Yaml;
use Drupal\ostickets\OSTicketsClientInterface;
use Drupal\webform\Plugin\WebformHandlerBase;
use Drupal\webform\WebformSubmissionConditionsValidatorInterface;
use Drupal\webform\WebformSubmissionInterface;
use Drupal\webform\WebformTokenManagerInterface;
use GuzzleHttp\Exception\GuzzleException;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Webform submission action handler.
 *
 * @WebformHandler(
 *   id = "ostickets",
 *   label = @Translation("OS Tickets: Create Ticket"),
 *   category = @Translation("Action"),
 *   description = @Translation("POST a ticket upon submission."),
 *   cardinality = \Drupal\webform\Plugin\WebformHandlerInterface::CARDINALITY_UNLIMITED,
 *   results = \Drupal\webform\Plugin\WebformHandlerInterface::RESULTS_PROCESSED,
 *   submission = \Drupal\webform\Plugin\WebformHandlerInterface::SUBMISSION_OPTIONAL,
 *   tokens = TRUE,
 * )
 */
class OSTicketsWebformHandler extends WebformHandlerBase {

  const BASE_FIELDS = [
    'email' => ['label' => 'Email', 'allowed_types' => NULL, 'allow_custom' => TRUE],
    'name' => ['label' => 'Name', 'allowed_types' => NULL, 'allow_custom' => TRUE],
    'subject' => ['label' => 'Subject', 'allowed_types' => NULL, 'allow_custom' => TRUE],
    'message' => ['label' => 'Message', 'allowed_types' => NULL, 'allow_custom' => TRUE],
  ];
  const ADDITONAL_FIELDS = [
    'priority' => ['label' => 'Priority', 'allowed_types' => ['select'], 'allow_custom' => TRUE],
    'attachments' => ['label' => 'Attachments', 'allowed_types' => ['managed_file', 'webform_image_file', 'webform_document_file']],
  ];

  /**
   * The webform token manager.
   *
   * @var \Drupal\webform\WebformTokenManagerInterface
   */
  protected $tokenManager;

  /**
   * The OS Tickets client
   *
   * @var \Drupal\ostickets\OSTicketsClientInterface
   */
  protected $osTicketsClient;

  /**
   * The entity type manager
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The file system
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected $fileSystem;

  /**
   * The current weight
   *
   * @var int
   */
  protected $currentWeight = 0;

  /**
   * {@inheritdoc}
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, LoggerChannelFactoryInterface $logger_factory, ConfigFactoryInterface $config_factory, EntityTypeManagerInterface $entity_type_manager, WebformSubmissionConditionsValidatorInterface $conditions_validator, WebformTokenManagerInterface $token_manager, OSTicketsClientInterface $ostickets_client, FileSystemInterface $file_system) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $logger_factory, $config_factory, $entity_type_manager, $conditions_validator);
    $this->tokenManager = $token_manager;
    $this->osTicketsClient = $ostickets_client;
    $this->entityTypeManager = $entity_type_manager;
    $this->fileSystem = $file_system;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('logger.factory'),
      $container->get('config.factory'),
      $container->get('entity_type.manager'),
      $container->get('webform_submission.conditions_validator'),
      $container->get('webform.token_manager'),
      $container->get('ostickets.client'),
      $container->get('file_system')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getSummary() {
    $configuration = $this->getConfiguration();
    $settings = $configuration['settings'];

    return [
      '#settings' => $settings,
    ] + parent::getSummary();
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    $settings = [
      'hostname' => '',
      'api_key' => '',
      'secure' => TRUE,
      'custom_fields' => [],
    ];

    $all_fields = self::BASE_FIELDS + self::ADDITONAL_FIELDS;

    foreach ($all_fields as $key => $field) {
      $settings["{$key}_field"] = '';
      if ($field['allow_custom']) {
        $settings["{$key}_custom"] = '';
      }
    }

    return $settings;
  }

  /**
   * Gets a configuration by key
   *
   * @param string $key
   *
   * @return mixed
   */
  private function getConfigurationKey($key) {
    $configuration = $this->getConfiguration()['settings'];
    return isset($configuration[$key]) ? $configuration[$key] : NULL;
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $this->elementTokenValidate($form);

    $form['hostname'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Hostname'),
      '#default_value' => $this->getConfigurationKey('hostname'),
    ];

    $form['api_key'] = [
      '#type' => 'textfield',
      '#title' => $this->t('API Key'),
      '#default_value' => $this->getConfigurationKey('api_key'),
    ];

    $form['secure'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Secure (Use HTTPS)'),
      '#default_value' => $this->getConfigurationKey('secure'),
    ];

    $all_fields = self::BASE_FIELDS + self::ADDITONAL_FIELDS;
    foreach ($all_fields as $key => $field) {
      // Skip the 'attachments' field, we will build it in a differnet way.
      // if ($key == 'attachments') {
      //   continue;
      // }

      $this->buildElement($key, $field, $form, in_array($key, array_keys(self::BASE_FIELDS)));
    }

    return $this->setSettingsParents($form);
  }

  /**
   * Builds an element with options.
   *
   * @param string $key
   * @param string $label
   * @param array $form
   * @param boolean $required
   *
   * @return array
   */
  protected function buildElement($key, $field, &$form, $required = TRUE) {
    $elements = $this->webform->getElementsDecodedAndFlattened();

    $options = array_filter($elements, function ($element) use ($field) {
      if (empty($field['allowed_types'])) {
        return $element['#type'] != 'captcha';
      }

      if (is_array($field['allowed_types'])) {
        return in_array($element['#type'], $field['allowed_types']);
      }
    });

    $options = array_map(function ($element) {return $element['#title'];}, $options);

    if ($field['allow_custom']) {
      $options['_token'] = $this->t('- Use Token -');
      $options['_custom'] = $this->t('- Custom -');
    }

    $container = [
      '#type' => 'container',
      '#tree' => FALSE,
    ];
    $container["{$key}_field"] = [
      '#type' => 'select',
      '#title' => $field['label'],
      '#options' => $options,
      '#empty_option' => $required ? NULL : $this->t('- None -'),
      '#required' => $required,
      '#default_value' => $this->getConfigurationKey("{$key}_field"),
    ];

    if ($field['allow_custom']) {
      $container["{$key}_custom"] = [
        '#type' => 'textfield',
        '#title' => $this->t("@label (Custom)", ['@label' => $field['label']]),
        '#states' => [
          'visible' => [
            "select[name='settings[{$key}_field]'" => [
              ['value' => '_token'],
              'or',
              ['value' => '_custom'],
            ],
          ],
        ],
        '#default_value' => $this->getConfigurationKey("{$key}_custom"),
      ];
    }

    $form["{$key}_container"] = $container;
  }

  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state) {
    if ($form_state->hasAnyErrors()) {
      return;
    }

    // Validate data element keys.
    $elements = $this->getWebform()->getElementsInitializedFlattenedAndHasValue();
    $data = Yaml::decode($form_state->getValue('data')) ?: [];
    foreach ($data as $key => $value) {
      if (!isset($elements[$key])) {
        $form_state->setErrorByName('data', $this->t('%key is not valid element key.', ['%key' => $key]));
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    parent::submitConfigurationForm($form, $form_state);
    $this->applyFormStateToConfiguration($form_state);

    // Cleanup states.
    $this->configuration['states'] = array_values(array_filter($this->configuration['states']));

    // Cleanup sticky.
    if ($form_state->getValue('sticky') === '') {
      $this->configuration['sticky'] = NULL;
    }

    // Cleanup locked.
    if ($form_state->getValue('locked') === '') {
      $this->configuration['locked'] = NULL;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function postSave(WebformSubmissionInterface $webform_submission, $update = TRUE) {
    $webform_data = $webform_submission->getData();
    $data_to_send = [];
    foreach (array_keys(self::BASE_FIELDS + self::ADDITONAL_FIELDS) as $key) {
      // Ignore the attachments field for now.
      if ($key == 'attachments') {
        continue;
      }

      // Get the field name.
      $field_name = $this->getConfigurationKey("{$key}_field");
      if ($field_name == '_token') {
        $value = $this->tokenManager->replace($this->getConfigurationKey("{$key}_custom"), $webform_submission);
        if (!empty($value)) {
          $data_to_send[$key] = $value;
        }
      } else if ($field_name == '_custom') {
        $value = $this->getConfigurationKey("{$key}_custom");
        if (!empty($value)) {
          $data_to_send[$key] = $value;
        }
      } else {
        if (!empty($webform_data[$field_name])) {
          $data_to_send[$key] = $webform_data[$field_name];
        }
      }
    }
    $attachments_key = $this->getConfigurationKey('attachments_field');
    if ($attachments_key && isset($webform_data[$attachments_key])) {
      if (is_int($webform_data[$attachments_key])) {
        $webform_data[$attachments_key] = [$webform_data[$attachments_key]];
      }

      foreach ($webform_data[$attachments_key] as $file_id) {
        /** @var \Drupal\file\FileInterface $file */
        $file = $this->entityTypeManager->getStorage('file')->load($file_id);
        $path = $this->fileSystem->realpath($file->getFileUri());
        $data = file_get_contents($path);
        $encoding = 'base64';
        $data = base64_encode($data);
        $data_to_send['attachments'][] = [$file->getFilename() => 'data:' . $file->getMimeType() . ';' . $encoding . ',' . mb_convert_encoding($data, 'UTF-8', 'UTF-8')];

      }
    }

    $this->osTicketsClient->setAPIKey($this->getConfigurationKey('api_key'));
    $this->osTicketsClient->setHostName($this->getConfigurationKey('hostname'));
    $this->osTicketsClient->setSecure($this->getConfigurationKey('secure'));
    try {
      $this->osTicketsClient->sendTicket($data_to_send);
    } catch (GuzzleException $e) {
      $this->messenger()->addError($e->getMessage());
    }
  }

}
